/*
 * nvgl2demo_common_kd.c
 *
 * Copyright 2009, NVIDIA Corporation.
 * All rights Reserved.
 *
 * THE INFORMATION CONTAINED HEREIN IS PROPRIETARY AND CONFIDENTIAL TO
 * NVIDIA, CORPORATION.  USE, REPRODUCTION OR DISCLOSURE TO ANY THIRD PARTY
 * IS SUBJECT TO WRITTEN PRE-APPROVAL BY NVIDIA, CORPORATION.
 */

//
// This file contains utility functions that we end up replicating in most
//   of our demos. It would be cleaner to have a single copy of this
//   somewhere, but replicating it makes the demos more self-contained for
//   teaching purposes.
//
#include <KD/kd.h>
#ifdef KD_NV_display
#include <KD/NV_display.h>
#endif
#include "nvgl2demo_common.h"
#include <stdio.h>

#ifdef _TEGRAPLATFORM
#include <KD/NV_extwindowprops.h>
#endif

#ifndef KD_DEFAULT_DISPLAY_NV
#define KD_DEFAULT_DISPLAY_NV 0
#endif

//
// Callback handling
//
static NvGl2DemoCloseCB   closeCB   = KD_NULL;
static NvGl2DemoResizeCB  resizeCB  = KD_NULL;
static NvGl2DemoKeyCB     keyCB     = KD_NULL;
static NvGl2DemoPointerCB pointerCB = KD_NULL;
static NvGl2DemoButtonCB  buttonCB  = KD_NULL;
static NvGl2DemoExtEventHandler extEventCB = KD_NULL;

struct NvGl2DemoPlatformState
{
    KDWindow* kdWindow;
};

#define MAX_ATTRIB 1  //maximum number of attributes passed to eglCreateWindowSurface
///////////////////////////////////////////////////////////////////////////////

//
// Argument parsing
//

// Extract an argument from the list
#if 0
static const char* NvGl2DemoArgExtract(int* argc, char** argv, int index)
{

	char* val = argv[index];
    int   j;
    for (j=index+1; j<*argc; ++j) {
        argv[j-1] = argv[j];
    }
    (*argc)--;
    return val;
}
#endif


// locate a substring in a null terminated string, return NULL if not found
static char* NvGl2DemoStrStr(const char* str1, const char* str2)
{
    int len1;
    int len2;
    KDchar* start;

    if (!str1 || !str2)
    {
        return KD_NULL;
    }

    len1 = kdStrlen(str1);
    len2 = kdStrlen(str2);

    // Find the first occurence
    start = kdStrchr(str1, *str2);

    // If we found the first character, continue to test
    // See if is possible to fit str2 into the rest of str1
    while ( start ) {

        if (!kdStrncmp(str2, start, len2)) {
            return start;
        }

        // Find the first occurence
        start = kdStrchr(start + 1, *str2);
    }

    // If we got here, we didn't find the string
    return KD_NULL;
}


#ifndef _TEGRAPLATFORM
// Temporary workaround until display driver can perform a proper clean up
// upon a control-C.  bug 604387.
//#include <signal.h>
#include <stdlib.h>
static void sig_int( int x )
{
    x;

    if (closeCB) {
        closeCB();
    } else {
        exit(-1);
    }
}
#endif
///////////////////////////////////////////////////////////////////////////////

//
// Starting up and shutting down window system and EGL
//

// egl/window system state
NvGl2DemoGfxState gfxState =
{
    (NativeDisplayType)0,
    (NativeWindowType)0,
    EGL_NO_DISPLAY,
    EGL_NO_SURFACE,
    (EGLConfig)0,
    EGL_NO_CONTEXT,
    0,
    0,
    2,
    0,
    KD_NULL,
};

// return gl2demo usage string under kd
char * NvGl2DemoGetUsageString(void)
{
  return "  [-res <xres> <yres>]\n \
 [-desktopsize <xres> <yres>]\n \
 [-screensize   <xres> <yres>]\n \
 [-windowsize  <xres> <yres>]\n \
 [-windowoffset <xres> <yres>]\n \
    (-windowoffset sets window property for a KDGUI-based window manager.)\n \
 [-disp <vga|dvi|lvds|tftlcd|clilcd|dsi|hdmi[#]>]\n \
    (only a subset of the -disp options are available on any given platform)\n \
 [-vcaa <samples (5 supported on Tegra)> ]\n";
}

// Start up, initializing native window system and EGL
int NvGl2DemoInitialize(int* argc, char** argv, const char *appName,
                        EGLint eglApiBit)
{
    // NOTE: This array is modified by the code below. The indices used
    // are hardcoded numbers, so be careful if you change the order of
    // the values!
    EGLint configAttrs[] = {
        EGL_RED_SIZE,        1,
        EGL_GREEN_SIZE,      1,
        EGL_BLUE_SIZE,       1,
        EGL_DEPTH_SIZE,     16,
        EGL_RENDERABLE_TYPE, EGL_NONE,  // NOTE: value gets overwritten
#ifdef EGL_NV_coverage_sample
        EGL_COVERAGE_BUFFERS_NV,  0,
        EGL_COVERAGE_SAMPLES_NV,  0,
#endif
        EGL_NONE
    };
    EGLint contextAttrs[] = {
        EGL_NONE, 0,                    // NOTE: value gets overwritten
                                        // for GL
        EGL_NONE
    };
    EGLenum eglApi;

#   ifdef KD_NV_display
    KDDesktopNV     *kdDesktop = KD_NULL;
    KDDisplayNV     *kdDisplay = KD_NULL;
    KDDisplayModeNV  mode;
    KDint            windowOffset[2] = { 0, 0};
    KDboolean        offsetSpecified = KD_FALSE;
	KDboolean		bool;
	KDchar*			displayName = KD_DEFAULT_DISPLAY_NV;
	KDint32			power;
#   endif
    KDint      desktopSize[2] = { 0, 0 };
    KDint      displaySize[2] = { 0, 0 };
    KDint      windowSize[2]  = { 0, 0 };
    EGLConfig* configList = KD_NULL;
    EGLint     configCount;
    EGLBoolean eglStatus;
    EGLint     vcaa = 0, buffering = 0;
    EGLint     windowAttrs[2*MAX_ATTRIB + 1];
    EGLint     windowIndex = 0;  //index of first kdFree place in windowAttrs
    int        i;
    const char*      query;
    KDchar		errorMessage[256];
    EGLint      desiredPixelSize = 16;
	long		theConfigIndex = 0;
    EGLint		thePixelDepth = 0;
	KDboolean	useOverlay = KD_FALSE;

#ifndef _TEGRAPLATFORM
    // Temporary workaround until display driver can perform a proper clean up
    // upon a control-C.  bug 604387.
    //signal(SIGINT, sig_int);
    // We also need to catch kill or killall.
    //signal(SIGTERM, sig_int);
#endif

    // Validate input API parameter and translate it to API enumerant
    switch (eglApiBit) {
        case EGL_OPENGL_ES2_BIT:
            eglApi = EGL_OPENGL_ES_API;
            break;

        case EGL_OPENVG_BIT:
            eglApi = EGL_OPENVG_API;
            break;

        default:
            kdLogMessage("Invalid/unsupported API requested\n");
            goto fail;
    }

    // Note: Use of parameters which modify the display configuration (-res,
    //       -screensize, -desktopsize, and -disp) is primarily provided when
    //       running the application in a standalone fashion. When there is
    //       a separate display server running, their use is discouraged, as
    //       they will blindly change the state without regard to any other
    //       applications sharing the display. Most openkode rendering
    //       applicatins will only need to set a window size and perhaps an
    //       offset. For applications which do have a need to adjust the
    //       display system, see the OpenKODE NV_display extension demos for
    //       better examples of how to perform various tasks.

    // Parse input parameters
    for (i=1; i<*argc; /*nop*/) {
        if (!kdStrncmp(argv[i], "-vcaa", 5)) {
            //argv[i] += 5;
            i++;
            if (i == *argc) {
                kdLogMessage("Missing parameter for -vcaa\n");
                goto fail;
            }
            vcaa = kdStrtol(argv[i], KD_NULL, 10);
            if (vcaa != 5) {
                kdStrcpy_s(errorMessage, 256, "Invalid parameter(");
                kdStrncat_s(errorMessage, 256, argv[i], 200);
                kdStrncat_s(errorMessage, 256, ") for -vcaa", 11);
                kdLogMessage(errorMessage);
                goto fail;
            }
            i++;
        }
        else if (!kdStrncmp(argv[i],"-buffering",10)) {
            //argv[i] += 10;
            i++;
            //if (argv[i][0] == 0) i++;
            if (i == *argc) {
                break;
            }

            buffering = kdStrtol(argv[i],KD_NULL,10);
            if(buffering < 0 || buffering > 4) {
                kdStrcpy_s(errorMessage, 256, "Invalid parameter(");
                kdStrncat_s(errorMessage, 256, argv[i], 200);
                kdStrncat_s(errorMessage, 256, ") for -buffering", 16);
                kdLogMessage(errorMessage);
                goto fail;
            }
            i++;
        }
        else if (!kdStrncmp(argv[i],"-res",4)) {
            //argv[i] += 4;
            i++;
            if (i == *argc) {
                kdLogMessage("Missing parameters for -res."
                             " Need to follow with xres yres.\n");
                goto fail;
            }

            desktopSize[0] = displaySize[0] = windowSize[0] =
                kdStrtol(argv[i],KD_NULL,10);
            i++;
            desktopSize[1] = displaySize[1] = windowSize[1] =
                kdStrtol(argv[i],KD_NULL,10);
            i++;
        }
#       ifdef KD_NV_display
        else if (!kdStrncmp(argv[i],"-desktopsize",12)) {
            //argv[i] += 12;
            i++;
            if (i == *argc) {
                kdLogMessage("Missing parameters for -desktopsize."
                             " Need to follow with xres yres.\n");
                goto fail;
            }

            desktopSize[0] = kdStrtol(argv[i],KD_NULL,10);
            i++;
            desktopSize[1] = kdStrtol(argv[i],KD_NULL,10);
            i++;
        }
        else if (!kdStrncmp(argv[i],"-screensize",11)) {
            //argv[i] += 11;
            i++;
            if (i == *argc) {
                kdLogMessage("Missing parameters for -screensize."
                             " Need to follow with xres yres.\n");
                goto fail;
            }

            displaySize[0] = kdStrtol(argv[i],KD_NULL,10);
            i++;
            displaySize[1] = kdStrtol(argv[i],KD_NULL,10);
            i++;
        }
        else if (!kdStrncmp(argv[i],"-windowsize",11)) {
            //argv[i] += 11;
            i++;
            if (i == *argc) {
                kdLogMessage("Missing parameters for -windowsize."
                             " Need to follow with xres yres.\n");
                goto fail;
            }

            windowSize[0] = kdStrtol(argv[i],KD_NULL,10);
            i++;
            windowSize[1] = kdStrtol(argv[i],KD_NULL,10);
            i++;
        }
        else if (!kdStrncmp(argv[i],"-windowoffset",13)) {
            //argv[i] += 13;
            i++;
            if (i == *argc) {
                kdLogMessage("Missing parameters for -windowoffset."
                             " Need to follow with xres yres.\n");
                goto fail;
            }

            windowOffset[0] = kdStrtol(argv[i],KD_NULL,10);
            i++;
            windowOffset[1] = kdStrtol(argv[i],KD_NULL,10);
            i++;
            offsetSpecified = KD_TRUE;
        }
        else if (!kdStrncmp(argv[i], "-disp", 5)) {
            i++;
            if (i == *argc) {
                kdLogMessage("Missing parameters for -disp."
                             " Need to follow with {vga|dvi|lvds|tftlcd|clilcd|dsi|hdmi}[#].\n");
                goto fail;
            }
            if (!kdStrcmp(argv[i], "vga")) {
                displayName = "Tegra:VGA0";
            } else if (!kdStrcmp(argv[i], "vga0")) {
                displayName = "Tegra:VGA0";
            } else if (!kdStrcmp(argv[i], "vga1")) {
                displayName = "Tegra:VGA1";
            } else if (!kdStrcmp(argv[i], "dvi")) {
                displayName = "Tegra:DVI0";
            } else if (!kdStrcmp(argv[i], "dvi0")) {
                displayName = "Tegra:DVI0";
            } else if (!kdStrcmp(argv[i], "dvi1")) {
                displayName = "Tegra:DVI1";
            } else if (!kdStrcmp(argv[i], "lvds")) {
                displayName = "Tegra:LVDS0";
            } else if (!kdStrcmp(argv[i], "lvds0")) {
                displayName = "Tegra:LVDS0";
            } else if (!kdStrcmp(argv[i], "lvds1")) {
                displayName = "Tegra:LVDS1";
            } else if (!kdStrcmp(argv[i], "tftlcd")) {
                displayName = "Tegra:TFTLCD0";
            } else if (!kdStrcmp(argv[i], "tftlcd0")) {
                displayName = "Tegra:TFTLCD0";
            } else if (!kdStrcmp(argv[i], "tftlcd1")) {
                displayName = "Tegra:TFTLCD1";
            } else if (!kdStrcmp(argv[i], "clilcd")) {
                displayName = "Tegra:CLILCD0";
            } else if (!kdStrcmp(argv[i], "clilcd0")) {
                displayName = "Tegra:CLILCD0";
            } else if (!kdStrcmp(argv[i], "clilcd1")) {
                displayName = "Tegra:CLILCD1";
            } else if (!kdStrcmp(argv[i], "dsi")) {
                displayName = "Tegra:DSI0";
            } else if (!kdStrcmp(argv[i], "dsi0")) {
                displayName = "Tegra:DSI0";
            } else if (!kdStrcmp(argv[i], "dsi1")) {
                displayName = "Tegra:DSI1";
            } else if (!kdStrcmp(argv[i], "hdmi")) {
                displayName = "Tegra:HDMI0";
            } else if (!kdStrcmp(argv[i], "hdmi0")) {
                displayName = "Tegra:HDMI0";
            } else if (!kdStrcmp(argv[i], "hdmi1")) {
                displayName = "Tegra:HDMI1";
            } else {
                kdLogMessage("Unknown parameter for -disp.");
                goto fail;
            }
            i++;
        }
#       endif
        else if (!kdStrncmp(argv[i],"-pixelsize",10)) {
            //argv[i] += 10;
            i++;
            if (i == *argc) {
                 kdLogMessage("Missing parameters for -pixelsize."
                             " Need to follow with {16|32}.\n");
                goto fail;
            }

            desiredPixelSize = kdStrtol(argv[i],KD_NULL,10);

            if(desiredPixelSize != 16 && desiredPixelSize != 32) {
                kdStrcpy_s(errorMessage, 256, "Invalid parameter(");
                kdStrncat_s(errorMessage, 256, argv[i], 200);
                kdStrncat_s(errorMessage, 256, ") for -pixelsize", 16);
                kdLogMessage(errorMessage);
                goto fail;
            }
            i++;
        }
		else if(!kdStrncmp(argv[i],"-overlay",8)) {
			//argv[i] += 8;
			i++;
			useOverlay=KD_TRUE;
		}
        else {
            i++;
        }
    }

    gfxState.platformspecific = (NvGl2DemoPlatformState*)kdMalloc(sizeof(NvGl2DemoPlatformState));
    if (!gfxState.platformspecific) {
        kdLogMessage("Could not allocate platform specific storage memory\n");
        goto fail;
    }
    gfxState.platformspecific->kdWindow = KD_NULL;

#   ifdef KD_NV_display
    // Get the default desktop
    kdDesktop = kdGetDesktopNV(KD_DEFAULT_DESKTOP_NV, KD_NULL);
    if (!kdDesktop) {
        kdLogMessage("Could not obtain KDDesktopNV pointer\n");
        goto fail;
    }

    // Get the requested display, enable it, and map to desktop
    kdDisplay = kdGetDisplayNV(displayName, KD_NULL);
    if (!kdDisplay) {
        kdLogMessage("Could not obtain KDDisplayNV pointer\n");
        goto fail;
    }
    bool = KD_TRUE;
    kdSetDisplayPropertybvNV(kdDisplay,
                             KD_DISPLAYPROPERTY_ENABLED_NV,
                             &bool);
    power = KD_DISPLAY_POWER_ON;
    kdSetDisplayPropertyivNV(kdDisplay,
                             KD_DISPLAYPROPERTY_POWER_NV,
                             &power);
    kdSetDisplayPropertycvNV(kdDisplay,
                             KD_DISPLAYPROPERTY_DESKTOP_NAME_NV,
                             KD_DEFAULT_DESKTOP_NV);

    // If a display size is requested, set it
    if (displaySize[0] && displaySize[1]) {
        mode.width   = displaySize[0];
        mode.height  = displaySize[1];
        mode.refresh = 60;
        if (kdSetDisplayModeNV(kdDisplay, &mode,
                               KD_DISPLAY_PROTOCOL_AUTOMATIC_NV)) {
            kdLogMessage("Could not set display mode\n");
            goto fail;
        }
    }

    // If a desktop size is requested, set it
    if (desktopSize[0] && desktopSize[1]) {
        if (kdSetDesktopPropertyivNV(kdDesktop, KD_DESKTOPPROPERTY_SIZE_NV,
                                     desktopSize)) {
            kdLogMessage("Could not set desktop size\n");
            goto fail;
        }
    }

    // Once we've set up the desktop and display we don't need them anymore
    kdReleaseDisplayNV(kdDisplay);
    kdReleaseDesktopNV(kdDesktop);
#   endif

    // Obtain the EGL display
    gfxState.display = eglGetDisplay(EGL_DEFAULT_DISPLAY);
    if (gfxState.display == EGL_NO_DISPLAY) {
        kdLogMessage("EGL failed to obtain display\n");
        goto fail;
    }

    // Initialize EGL
    eglStatus = eglInitialize(gfxState.display, 0, 0);
    if (!eglStatus) {
        kdLogMessage("EGL failed to initialize\n");
        goto fail;
    }

    // If VCAA requested and supported, get a config that supports it
    if (vcaa > 1) {
        query = eglQueryString(gfxState.display, EGL_EXTENSIONS);
        if(!NvGl2DemoStrStr(query, " EGL_NV_coverage_sample")) {
            kdLogMessage("VCAA not supported!\n");
            goto fail;
        }
        configAttrs[11]  = 1;
        configAttrs[13]  = vcaa;
    }

    // Put API-specific values into the config and context attributes.
    // This assumes that the EGL_RENDERABLE_TYPE enum is the fourth one
    // in the configAttrs array and that the contextAttrs array is
    // empty but has room for an EGL_CONTEXT_CLIENT_VERSION enum.
    //
    // Also, while we're at it, bind the client API.
    configAttrs[9] = eglApiBit;
    switch (eglApiBit) {
        case EGL_OPENGL_ES2_BIT:
            contextAttrs[0] = EGL_CONTEXT_CLIENT_VERSION;
            contextAttrs[1] = 2;
            break;

        default:
            break;
    }
    eglBindAPI(eglApi);

    // Find out how many configurations suit our needs
    eglStatus = eglChooseConfig(gfxState.display, configAttrs,
                                KD_NULL, 0, &configCount);
    if (!eglStatus || !configCount) {
        kdLogMessage("EGL failed to return any matching configurations\n");
        goto fail;
    }

    // Allocate room for the list of matching configurations
    configList = (EGLConfig*)kdMalloc(configCount * sizeof(EGLConfig));
    if (!configList) {
        kdLogMessage("kdMalloc failure obtaining configuration list\n");
        goto fail;
    }

    // Obtain the configuration list from EGL
    eglStatus = eglChooseConfig(gfxState.display, configAttrs,
                                configList, configCount, &configCount);
    if (!eglStatus || !configCount) {
        kdLogMessage("EGL failed to populate configuration list\n");
        goto fail;
    }

    // Select an EGL configuration that matches the native window and desired pixel size
    for( ; theConfigIndex < configCount; ++theConfigIndex )
    {
        eglGetConfigAttrib( gfxState.display, configList[theConfigIndex], EGL_BUFFER_SIZE, &thePixelDepth );
        if ( thePixelDepth == desiredPixelSize )
	{        	
            gfxState.config = configList[theConfigIndex];
            break;
        }
    }

    if ( thePixelDepth != desiredPixelSize )
    {
        kdLogMessage( "Unable to set desired pixel size\n" );
        goto fail;
    }
    else
    {
    	EGLint theRedSize = 0;
        EGLint theGreenSize = 0;
        EGLint theBlueSize = 0;
        EGLint theAlphaSize = 0;
        char theMessageBuffer[256];

	eglGetConfigAttrib( gfxState.display, configList[theConfigIndex], EGL_RED_SIZE, &theRedSize );
       	eglGetConfigAttrib( gfxState.display, configList[theConfigIndex], EGL_GREEN_SIZE, &theGreenSize );
	eglGetConfigAttrib( gfxState.display, configList[theConfigIndex], EGL_BLUE_SIZE, &theBlueSize );
	eglGetConfigAttrib( gfxState.display, configList[theConfigIndex], EGL_ALPHA_SIZE, &theAlphaSize );
	
	if ( theAlphaSize > 0 )
    	    sprintf( theMessageBuffer, "\nDisplay set to RGBA%d%d%d%d (%dbit)", theRedSize, theGreenSize, theBlueSize, theAlphaSize, thePixelDepth );
        else
    	    sprintf( theMessageBuffer, "\nDisplay set to RGB%d%d%d (%dbit)", theRedSize, theGreenSize, theBlueSize, thePixelDepth );

	kdLogMessage( theMessageBuffer );
    }	

    // Set buffering options
    if(buffering) {

        gfxState.buffering = buffering;

        query = eglQueryString(gfxState.display, EGL_EXTENSIONS);
        windowAttrs[windowIndex++] = EGL_RENDER_BUFFER;

        switch(buffering) {
            case 1:
                windowAttrs[windowIndex++] = EGL_SINGLE_BUFFER;
                break;

            case 2:
                windowAttrs[windowIndex++] = EGL_BACK_BUFFER;
                break;

            #ifdef EGL_TRIPLE_BUFFER_NV
            case 3:
                if(!NvGl2DemoStrStr(query, "EGL_NV_triple_buffer ")) {
                    kdLogMessage("TRIPLE_BUFFER not supported!\n");
                    goto fail;
                }
                windowAttrs[windowIndex++] = EGL_TRIPLE_BUFFER_NV;
                break;
            #endif

            #ifdef EGL_QUADRUPLE_BUFFER_NV
            case 4:
                if(!NvGl2DemoStrStr(query, "EGL_NV_quadruple_buffer ")) {
                    kdLogMessage("QUADRUPLE_BUFFER not supported!\n");
                    goto fail;
                }
                windowAttrs[windowIndex++] = EGL_QUADRUPLE_BUFFER_NV;
                break;
            #endif

            default:
                kdLogMessage("Choosen buffering not defined.\n");
                goto fail;
        }
    }

    // Create a native window
    gfxState.platformspecific->kdWindow = kdCreateWindow(gfxState.display, gfxState.config, KD_NULL);
    if (!gfxState.platformspecific->kdWindow) {
        kdLogMessage("Error creating native window\n");
        goto fail;
    }

	#if defined (_PCPLATFORM) || defined (_TEGRAPLATFORM)
		// Set window caption
		kdSetWindowPropertycv( gfxState.platformspecific->kdWindow, KD_WINDOWPROPERTY_CAPTION, appName );
	#endif

#if defined (_TEGRAPLATFORM)
	if (!windowSize[0])
	{
		KDboolean fullscreen = KD_TRUE;
		kdSetWindowPropertybv(gfxState.platformspecific->kdWindow, KD_WINDOWPROPERTY_FULLSCREEN_NV, &fullscreen);
	}

	if(useOverlay)
	{
		KDboolean overlay = KD_TRUE;
		kdSetWindowPropertybv(gfxState.platformspecific->kdWindow, KD_WINDOWPROPERTY_OVERLAY_NV, &overlay);
	}
#else
    // Set window size. If no specified, default to 800x480
    if (!windowSize[0]) windowSize[0] = 800;
    if (!windowSize[1]) windowSize[1] = 480;
#endif
#if 1
    if (kdSetWindowPropertyiv(gfxState.platformspecific->kdWindow,
                          KD_WINDOWPROPERTY_SIZE, windowSize))
    {
        kdLogMessage("Could not set native window size\n");
        goto fail;
    }
#endif

#ifdef KD_NV_display
    if (offsetSpecified) {
        if (kdSetWindowPropertyiv(gfxState.platformspecific->kdWindow,
                           KD_WINDOWPROPERTY_DESKTOP_OFFSET_NV, windowOffset))
        {
            kdLogMessage("Could not set desktop offset");
            goto fail;
        }
    }
#endif

    if (kdRealizeWindow(gfxState.platformspecific->kdWindow, &gfxState.nativeWindow)) {
        kdLogMessage("Could not realize native window\n");
        goto fail;
    }

    // Create an EGL window surface for the native window
    windowAttrs[windowIndex++] = EGL_NONE;
    gfxState.surface = eglCreateWindowSurface(gfxState.display,
                                              gfxState.config,
                                              gfxState.nativeWindow,
                                              windowAttrs);
    if (!gfxState.surface) {
        kdLogMessage("EGL couldn't create window\n");
        goto fail;
    }

    // Create an EGL context
    gfxState.context = eglCreateContext(gfxState.display,
                                        gfxState.config,
                                        KD_NULL,
                                        contextAttrs);
    if (!gfxState.context) {
        kdLogMessage("EGL couldn't create context\n");
        goto fail;
    }

    // Make the context and surface current for rendering
    eglStatus = eglMakeCurrent(gfxState.display,
                               gfxState.surface, gfxState.surface,
                               gfxState.context);
    if (!eglStatus) {
        kdLogMessage("EGL couldn't make context/surface current\n");
        goto fail;
    }

    // Store the rendering API away
    gfxState.api = eglApi;

    // Query the EGL surface width and height
    eglStatus =  eglQuerySurface(gfxState.display, gfxState.surface,
                                 EGL_WIDTH,  &gfxState.width)
              && eglQuerySurface(gfxState.display, gfxState.surface,
                                 EGL_HEIGHT, &gfxState.height);
    if (!eglStatus) {
        kdLogMessage("EGL couldn't get window width/height\n");
        goto fail;
    }

    // kdFree the configuration list and return success
    kdFree(configList);
    return 1;

    // On failure, clean up partial initialization
    fail:
    if (configList) kdFree(configList);
    NvGl2DemoShutdown();
    return 0;
}

// Shut down, kdFreeing all EGL and native window system resources.
void NvGl2DemoShutdown(void)
{
    EGLBoolean eglStatus;

    // Clear rendering context
    // Note that we need to bind the API to unbind... yick
    if (gfxState.display != EGL_NO_DISPLAY) {
        eglBindAPI(gfxState.api);
        eglStatus = eglMakeCurrent(gfxState.display,
                                   EGL_NO_SURFACE, EGL_NO_SURFACE,
                                   EGL_NO_CONTEXT);
        if (!eglStatus)
            kdLogMessage("Error clearing current surfaces/context\n");
    }

    // Destroy the EGL context
    if (gfxState.context != EGL_NO_CONTEXT) {
        eglStatus = eglDestroyContext(gfxState.display, gfxState.context);
        if (!eglStatus)
            kdLogMessage("Error destroying EGL context\n");
        gfxState.context = EGL_NO_CONTEXT;
    }

    // Destroy the EGL surface
    if (gfxState.surface != EGL_NO_SURFACE) {
        eglStatus = eglDestroySurface(gfxState.display, gfxState.surface);
        if (!eglStatus)
            kdLogMessage("Error destroying EGL surface\n");
        gfxState.surface = EGL_NO_SURFACE;
    }

    // Terminate EGL
    if (gfxState.display != EGL_NO_DISPLAY) {
        eglStatus = eglTerminate(gfxState.display);
        if (!eglStatus)
            kdLogMessage("Error terminating EGL\n");
        gfxState.display = EGL_NO_DISPLAY;
    }

    // Close the native window
    if (gfxState.platformspecific && gfxState.platformspecific->kdWindow) {
        if ( kdDestroyWindow(gfxState.platformspecific->kdWindow) ) {
            kdLogMessage("Error destroying native window\n");
        }
        gfxState.platformspecific->kdWindow = KD_NULL;
        gfxState.nativeWindow = (NativeWindowType)0;
        kdFree(gfxState.platformspecific);
        gfxState.platformspecific = KD_NULL;
    }
}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

//
// Callback handling
//
void NvGl2DemoSetCloseCB(NvGl2DemoCloseCB cb) { closeCB  = cb; }
void NvGl2DemoSetResizeCB(NvGl2DemoResizeCB cb) { resizeCB = cb; }
void NvGl2DemoSetKeyCB(NvGl2DemoKeyCB cb) { keyCB = cb; }
void NvGl2DemoSetPointerCB(NvGl2DemoPointerCB cb) { pointerCB = cb; }
void NvGl2DemoSetButtonCB(NvGl2DemoButtonCB cb) { buttonCB = cb; }
void NvGl2DemoSetExtEventHandler(NvGl2DemoExtEventHandler cb){ extEventCB = cb; }

void NvGl2DemoCheckEvents(void)
{
    const KDEvent *event;

#ifdef _PCPLATFORM
	static KDust theWaitTime = NV_WETIME_60FPS;
#else
	static KDust theWaitTime = NV_WETIME_NEVER;
#endif

    while ((event = kdWaitEvent(theWaitTime)) != KD_NULL) {
        switch (event->type) {

            case KD_EVENT_WINDOW_CLOSE:
                if (closeCB)
                    closeCB();
                break;

            case KD_EVENT_WINDOWPROPERTY_CHANGE:
                if (event->data.windowproperty.pname==KD_WINDOWPROPERTY_SIZE) {
                    KDint32 size[2];
                    kdGetWindowPropertyiv(gfxState.platformspecific->kdWindow,
                                          KD_WINDOWPROPERTY_SIZE, size);
                    if (resizeCB)
                        resizeCB(size[0], size[1]);
                } else {
                    kdDefaultEvent(event);
                }
                break;

#if 0
            // NOTE: This uses a non standardized API for text input.
            //       No guarantee is made that this will continue to work.
            case KD_EVENT_INPUT_TEXT_NV:
                if (keyCB)
                    keyCB(event->data.user.value1.i32pair.a, 1);
                break;
#endif

            case KD_EVENT_INPUT_POINTER:
                if (pointerCB)
                    pointerCB(event->data.inputpointer.x,
                              event->data.inputpointer.y);
                if (buttonCB &&
                    (event->data.inputpointer.index==KD_INPUT_POINTER_SELECT))
                    buttonCB(1, event->data.inputpointer.select);
                break;
			
			#ifdef _PCPLATFORM	
				case KD_EVENT_WINDOW_FOCUS:
					{
						KDboolean theHasFocus;
						if ( gfxState.platformspecific->kdWindow )
						{
							kdGetWindowPropertybv( gfxState.platformspecific->kdWindow, KD_WINDOWPROPERTY_FOCUS, &theHasFocus );
						
							// Switch to NV_WETIME_1FPS so we can re-focus window properly (Windows)
							if ( theHasFocus == KD_TRUE )
								theWaitTime = NV_WETIME_NEVER;
							else
								theWaitTime = NV_WETIME_1FPS;
						}
					}
				break;
			#endif

            default:
                kdDefaultEvent(event);
        }

		if ( extEventCB )
			extEventCB( event );
    }
}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

//
// Shader program creation
//

// Loads a file containing shader code
static char* NvGl2DemoLoadFile(const char *file)
{
    char path[1024]; // This is reused for error messages as well
    KDStat st;
    char *data;
    KDFile *f;

    // Look in both /res and /data dir
    kdStrcpy_s(path, 1024, "/res/");
    kdStrncat_s(path, 1024, file, 1000);
    f = kdFopen(path, "rb");

    if (!f) {
        kdStrcpy_s(path, 1024, "/data/");
        kdStrncat_s(path, 1024, file, 1000);
        f = kdFopen(path, "rb");
    }

    if (!f) {
        kdStrcpy_s(path, 1024, "Unable to open shader: ");
        kdStrncat_s(path, 1024, file, 1000);
        kdLogMessage(path);
        return 0;
    }

    if(kdStat(path, &st)) {
        kdStrcpy_s(path, 1024, "Unable to stat shader: ");
        kdStrncat_s(path, 1024, file, 1000);
        kdLogMessage(path);
        kdFclose(f);
        return 0;
    }

    data = (char *)kdMalloc( (KDsize)st.st_size + 1 );
    if(!data) {
        kdStrcpy_s(path, 1024, "Unable to allocate memory for shader: ");
        kdStrncat_s(path, 1024, file, 1000);
        kdLogMessage(path);
        kdFclose(f);
        return 0;
    }
    if(kdFread(data, 1, (KDsize)st.st_size, f) != st.st_size) {
        kdStrcpy_s(path, 1024, "Unable to read shader: ");
        kdStrncat_s(path, 1024, file, 1000);
        kdLogMessage(path);

        kdFclose(f);
        return 0;
    }
    data[st.st_size] = 0;
    kdFclose(f);

    return data;
}

// If shader compilation fails, prints the associated logs
static void NvGl2DemoShaderDebug(GLuint obj, GLenum status, const char* op)
{
    KDchar errorMessage[1024];
    int success;
    // log output.
    int len;
    char *str = KD_NULL;
    if (status == GL_COMPILE_STATUS) {
        glGetShaderiv(obj, GL_INFO_LOG_LENGTH, &len);
        if (len > 0) {
            str = (char *)kdMalloc(len * sizeof(char));
            glGetShaderInfoLog(obj, len, KD_NULL, str);
        }
    } else { // LINK
        glGetProgramiv(obj, GL_INFO_LOG_LENGTH, &len);
        if (len > 0) {
            str = (char *)kdMalloc(len * sizeof(char));
            glGetProgramInfoLog(obj, len, KD_NULL, str);
        }
    }
    if (str != KD_NULL && *str != '\0') {
        kdStrcpy_s(errorMessage, 1024, "--- ");
        kdStrncat_s(errorMessage, 1024, op, 1000);
        kdStrncat_s(errorMessage, 1024, " log ---", 8);
        kdLogMessage(errorMessage);
        kdLogMessage(str);
    }
    if (str) { kdFree(str); }

    // check the compile / link status.
    if (status == GL_COMPILE_STATUS) {
        glGetShaderiv(obj, status, &success);
        if (!success) {
            glGetShaderiv(obj, GL_SHADER_SOURCE_LENGTH, &len);
            if (len > 0) {
                str = (char *)kdMalloc(len * sizeof(char));
                glGetShaderSource(obj, len, KD_NULL, str);
                if (str != KD_NULL && *str != '\0') {
                    kdStrcpy_s(errorMessage, 1024, "--- ");
                    kdStrncat_s(errorMessage, 1024, op, 1000);
                    kdStrncat_s(errorMessage, 1024, " code ---", 8);
                    kdLogMessage(errorMessage);
                    kdLogMessage(str);
                }
                kdFree(str);
            }
        }
    } else { // LINK
        glGetProgramiv(obj, status, &success);
    }

    if (!success)
    {
        kdStrcpy_s(errorMessage, 1024, "--- ");
        kdStrncat_s(errorMessage, 1024, op, 1000);
        kdStrncat_s(errorMessage, 1024, " failed", 7);
        kdLogMessage(errorMessage);
        kdExit(-1);
    }
}

// Take precompiled shader binaries and builds a shader program
GLuint NvGl2DemoLoadShaderBinaries(
    const char* vertBin, GLuint vertBinSize,
    const char* fragBin, GLuint fragBinSize,
    GLboolean debugging)
{
// Binary shaders not supportable for non-ES OpenGL
#ifdef GL_ES_VERSION_2_0

    GLuint prog;
    GLuint vertShader;
    GLuint fragShader;

    // Create the program
    prog = glCreateProgram();

    // Create the GL shader objects
    vertShader = glCreateShader(GL_VERTEX_SHADER);
    fragShader = glCreateShader(GL_FRAGMENT_SHADER);

    // Load the binary data into the shader objects
    glShaderBinary(1, &vertShader,
                   GL_NVIDIA_PLATFORM_BINARY_NV, vertBin, vertBinSize);
    glShaderBinary(1, &fragShader,
                   GL_NVIDIA_PLATFORM_BINARY_NV, fragBin, fragBinSize);

    // Attach the shaders to the program
    glAttachShader(prog, vertShader);
    glAttachShader(prog, fragShader);

#if 0
    // Delete the shaders
    glDeleteShader(vertShader);
    glDeleteShader(fragShader);
#endif

    // Link the shader program
    glLinkProgram(prog);
    if (debugging)
        NvGl2DemoShaderDebug(prog, GL_LINK_STATUS, "Program Link");

    return prog;
#else  // GL_ES_VERSION_2_0
    return 0;
#endif // GL_ES_VERSION_2_0
}

// Takes shader source files, compiles them, and builds a shader program
GLuint NvGl2DemoLoadShaderSources(
    const char* vertFile,
    const char* fragFile,
    GLboolean debugging)
{
    GLuint prog = 0;
    char*  vertSource;
    char*  fragSource;
    GLint  vertSourceLen;
    GLint  fragSourceLen;
    GLuint vertShader;
    GLuint fragShader;

    // Load the shader files
    vertSource    = NvGl2DemoLoadFile(vertFile);
    fragSource    = NvGl2DemoLoadFile(fragFile);
    if (!vertSource || !fragSource) goto done;
    vertSourceLen = (GLint)kdStrlen(vertSource);
    fragSourceLen = (GLint)kdStrlen(fragSource);

    // Create the program
    prog = glCreateProgram();

    // Create the GL shader objects
    vertShader = glCreateShader(GL_VERTEX_SHADER);
    fragShader = glCreateShader(GL_FRAGMENT_SHADER);

    // Load shader sources into GL and compile
    glShaderSource(vertShader, 1, (const char**)&vertSource, &vertSourceLen);
    glCompileShader(vertShader);
    if (debugging)
        NvGl2DemoShaderDebug(vertShader, GL_COMPILE_STATUS, "Vert Compile");
    glShaderSource(fragShader, 1, (const char**)&fragSource, &fragSourceLen);
    glCompileShader(fragShader);
    if (debugging)
        NvGl2DemoShaderDebug(fragShader, GL_COMPILE_STATUS, "Frag Compile");

    // Attach the shaders to the program
    glAttachShader(prog, vertShader);
    glAttachShader(prog, fragShader);

    // Delete the shaders
    glDeleteShader(vertShader);
    glDeleteShader(fragShader);

    // Link and validate the shader program
    glLinkProgram(prog);
    if (debugging)
        NvGl2DemoShaderDebug(prog, GL_LINK_STATUS, "Program Link");
    glValidateProgram(prog);
    if (debugging)
        NvGl2DemoShaderDebug(prog, GL_VALIDATE_STATUS, "Program Validate");

    done:

    kdFree(fragSource);
    kdFree(vertSource);
    return prog;
}

///////////////////////////////////////////////////////////////////////////////

// Math stuff

int eq(float a, float b)  {
    float diff = a-b;
    if (diff < 0) {
        diff = -diff;
    }
    return diff <= eps;
}

//
// Matrix functions, since GLES 2.x doesn't provide them
//


void matrix_identity(float m[16])
{
    kdMemset(m, 0, sizeof(float) * 16);
    m[4 * 0 + 0] = m[4 * 1 + 1] = m[4 * 2 + 2] = m[4 * 3 + 3] = 1.0;
}

int matrix_equals(float a[16], float b[16])  {
    int i;
    for(i = 0; i < 16; ++i)
        if(!eq(a[i], b[i]))
            return 0;

    return 1;
}

void matrix_transpose(float m[16])
{
    int i, j;
    float t;
    for(i = 1; i < 4; ++i)
        for(j = 0; j < i; ++j)  {
            t = m[4*i+j];
            m[4*i+j] = m[4*j+i];
            m[4*j+i] = t;
        }
}

void matrix_multiply(float m0[16], float m1[16])
{
    int r, c, i;
    for(r = 0; r < 4; r++)
    {
        float m[4] = {0.0, 0.0, 0.0, 0.0};
        for(c = 0; c < 4; c++)
        {
            for(i = 0; i < 4; i++)
            {
                m[c] += m0[4 * i + r] * m1[4 * c + i];
            }
        }
        for(c = 0; c < 4; c++) { m0[4 * c + r] = m[c]; }
    }
}

void matrix_multiply_4x4_3x3(float m0[16], float m1[9])
{
    int r, c, i;
    for(r = 0; r < 4; r++)
    {
        float m[3] = {0.0, 0.0, 0.0};
        for(c = 0; c < 3; c++)
        {
            for(i = 0; i < 3; i++)
            {
                m[c] += m0[4 * i + r] * m1[3 * c + i];
            }
        }
        for(c = 0; c < 3; c++) { m0[4 * c + r] = m[c]; }
    }
}

void matrix_multiply_3x3(float m0[9], float m1[9])
{
    int r, c, i;
    for(r = 0; r < 3; r++)
    {
        float m[3] = {0.0, 0.0, 0.0};
        for(c = 0; c < 3; c++)
        {
            for(i = 0; i < 3; i++)
            {
                m[c] += m0[3 * i + r] * m1[3 * c + i];
            }
        }
        for(c = 0; c < 3; c++) { m0[3 * c + r] = m[c]; }
    }
}

void matrix_frustum(float m[16],
                    float l, float r, float b, float t, float n, float f)
{
    float m1[16];
    float rightMinusLeftInv, topMinusBottomInv, farMinusNearInv, twoNear;

    rightMinusLeftInv = 1.0f / (r - l);
    topMinusBottomInv = 1.0f / (t - b);
    farMinusNearInv = 1.0f / (f - n);
    twoNear = 2.0f * n;

    m1[ 0] = twoNear * rightMinusLeftInv;
    m1[ 1] = 0.0f;
    m1[ 2] = 0.0f;
    m1[ 3] = 0.0f;

    m1[ 4] = 0.0f;
    m1[ 5] = twoNear * topMinusBottomInv;
    m1[ 6] = 0.0f;
    m1[ 7] = 0.0f;

    m1[ 8] = (r + l) * rightMinusLeftInv;
    m1[ 9] = (t + b) * topMinusBottomInv;
    m1[10] = -(f + n) * farMinusNearInv;
    m1[11] = -1.0f;

    m1[12] = 0.0f;
    m1[13] = 0.0f;
    m1[14] = -(twoNear * f) * farMinusNearInv;
    m1[15] = 0.0f;

    matrix_multiply(m, m1);
}

void matrix_ortho(float m[16],
                  float l, float r, float b, float t, float n, float f)
{
    float m1[16];
    float rightMinusLeftInv, topMinusBottomInv, farMinusNearInv;

    rightMinusLeftInv = 1.0f / (r - l);
    topMinusBottomInv = 1.0f / (t - b);
    farMinusNearInv = 1.0f / (f - n);

    m1[ 0] = 2.0f * rightMinusLeftInv;
    m1[ 1] = 0.0f;
    m1[ 2] = 0.0f;
    m1[ 3] = 0.0f;

    m1[ 4] = 0.0f;
    m1[ 5] = 2.0f * topMinusBottomInv;
    m1[ 6] = 0.0f;
    m1[ 7] = 0.0f;

    m1[ 8] = 0.0f;
    m1[ 9] = 0.0f;
    m1[10] = -2.0f * farMinusNearInv;
    m1[11] = 0.0f;

    m1[12] = -(r + l) * rightMinusLeftInv;
    m1[13] = -(t + b) * topMinusBottomInv;
    m1[14] = -(f + n) * farMinusNearInv;
    m1[15] = 1.0f;

    matrix_multiply(m, m1);
}

void matrix_translate(float m[16], float x, float y, float z)
{
    float m1[16];
    matrix_identity(m1);

    m1[4 * 3 + 0] = x;
    m1[4 * 3 + 1] = y;
    m1[4 * 3 + 2] = z;

    matrix_multiply(m, m1);
}

void matrix_rotate_create3x3(float m[9],
                             float theta, float x, float y, float z)
{
    float len = kdSqrtf(x * x + y * y + z * z);
    float u0 = x / len;
    float u1 = y / len;
    float u2 = z / len;
    float rad = (float)(theta / 180 * KD_PI_F);
    float c = kdCosf(rad);
    float s = kdSinf(rad);
    m[3 * 0 + 0] = u0 * u0 + c * (1 - u0 * u0) + s * 0;
    m[3 * 0 + 1] = u0 * u1 + c * (0 - u0 * u1) + s * u2;
    m[3 * 0 + 2] = u0 * u2 + c * (0 - u0 * u2) - s * u1;

    m[3 * 1 + 0] = u1 * u0 + c * (0 - u1 * u0) - s * u2;
    m[3 * 1 + 1] = u1 * u1 + c * (1 - u1 * u1) + s * 0;
    m[3 * 1 + 2] = u1 * u2 + c * (0 - u1 * u2) + s * u0;

    m[3 * 2 + 0] = u2 * u0 + c * (0 - u2 * u0) + s * u1;
    m[3 * 2 + 1] = u2 * u1 + c * (0 - u2 * u1) - s * u0;
    m[3 * 2 + 2] = u2 * u2 + c * (1 - u2 * u2) + s * 0;
}

void matrix_rotate(float m[16], float theta, float x, float y, float z)
{
    float r[9];
    matrix_rotate_create3x3(r, theta, x, y, z);
    matrix_multiply_4x4_3x3(m, r);
}

void matrix_rotate_3x3(float m[9], float theta, float x, float y, float z)
{
    float r[9];
    matrix_rotate_create3x3(r, theta, x, y, z);
    matrix_multiply_3x3(m, r);
}

float matrix_determinant(float m[16]) {
   return
       m[4*0+3] * m[4*1+2] * m[4*2+1] * m[4*3+0]-m[4*0+2] * m[4*1+3] * m[4*2+1] * m[4*3+0]-m[4*0+3] * m[4*1+1] * m[4*2+2] * m[4*3+0]+m[4*0+1] * m[4*1+3]    * m[4*2+2] * m[4*3+0]+
       m[4*0+2] * m[4*1+1] * m[4*2+3] * m[4*3+0]-m[4*0+1] * m[4*1+2] * m[4*2+3] * m[4*3+0]-m[4*0+3] * m[4*1+2] * m[4*2+0] * m[4*3+1]+m[4*0+2] * m[4*1+3]    * m[4*2+0] * m[4*3+1]+
       m[4*0+3] * m[4*1+0] * m[4*2+2] * m[4*3+1]-m[4*0+0] * m[4*1+3] * m[4*2+2] * m[4*3+1]-m[4*0+2] * m[4*1+0] * m[4*2+3] * m[4*3+1]+m[4*0+0] * m[4*1+2]    * m[4*2+3] * m[4*3+1]+
       m[4*0+3] * m[4*1+1] * m[4*2+0] * m[4*3+2]-m[4*0+1] * m[4*1+3] * m[4*2+0] * m[4*3+2]-m[4*0+3] * m[4*1+0] * m[4*2+1] * m[4*3+2]+m[4*0+0] * m[4*1+3]    * m[4*2+1] * m[4*3+2]+
       m[4*0+1] * m[4*1+0] * m[4*2+3] * m[4*3+2]-m[4*0+0] * m[4*1+1] * m[4*2+3] * m[4*3+2]-m[4*0+2] * m[4*1+1] * m[4*2+0] * m[4*3+3]+m[4*0+1] * m[4*1+2]    * m[4*2+0] * m[4*3+3]+
       m[4*0+2] * m[4*1+0] * m[4*2+1] * m[4*3+3]-m[4*0+0] * m[4*1+2] * m[4*2+1] * m[4*3+3]-m[4*0+1] * m[4*1+0] * m[4*2+2] * m[4*3+3]+m[4*0+0] * m[4*1+1]    * m[4*2+2] * m[4*3+3];
}

void matrix_inverse(float m[16]) {
    float a[16];
    float det;
    int i;
    float b[16], e[16];

    a[4*0+0] = m[4*1+2]*m[4*2+3]*m[4*3+1] - m[4*1+3]*m[4*2+2]*m[4*3+1] + m[4*1+3]*m[4*2+1]*m[4*3+2] - m[4*1+1]*m[4*2+3]*m[4*3+2] - m[4*1+2]*m[4*2+1]*m[4*3+3] + m[4*1+1]*m[4*2+2]*m[4*3+3];
    a[4*0+1] = m[4*0+3]*m[4*2+2]*m[4*3+1] - m[4*0+2]*m[4*2+3]*m[4*3+1] - m[4*0+3]*m[4*2+1]*m[4*3+2] + m[4*0+1]*m[4*2+3]*m[4*3+2] + m[4*0+2]*m[4*2+1]*m[4*3+3] - m[4*0+1]*m[4*2+2]*m[4*3+3];
    a[4*0+2] = m[4*0+2]*m[4*1+3]*m[4*3+1] - m[4*0+3]*m[4*1+2]*m[4*3+1] + m[4*0+3]*m[4*1+1]*m[4*3+2] - m[4*0+1]*m[4*1+3]*m[4*3+2] - m[4*0+2]*m[4*1+1]*m[4*3+3] + m[4*0+1]*m[4*1+2]*m[4*3+3];
    a[4*0+3] = m[4*0+3]*m[4*1+2]*m[4*2+1] - m[4*0+2]*m[4*1+3]*m[4*2+1] - m[4*0+3]*m[4*1+1]*m[4*2+2] + m[4*0+1]*m[4*1+3]*m[4*2+2] + m[4*0+2]*m[4*1+1]*m[4*2+3] - m[4*0+1]*m[4*1+2]*m[4*2+3];
    a[4*1+0] = m[4*1+3]*m[4*2+2]*m[4*3+0] - m[4*1+2]*m[4*2+3]*m[4*3+0] - m[4*1+3]*m[4*2+0]*m[4*3+2] + m[4*1+0]*m[4*2+3]*m[4*3+2] + m[4*1+2]*m[4*2+0]*m[4*3+3] - m[4*1+0]*m[4*2+2]*m[4*3+3];
    a[4*1+1] = m[4*0+2]*m[4*2+3]*m[4*3+0] - m[4*0+3]*m[4*2+2]*m[4*3+0] + m[4*0+3]*m[4*2+0]*m[4*3+2] - m[4*0+0]*m[4*2+3]*m[4*3+2] - m[4*0+2]*m[4*2+0]*m[4*3+3] + m[4*0+0]*m[4*2+2]*m[4*3+3];
    a[4*1+2] = m[4*0+3]*m[4*1+2]*m[4*3+0] - m[4*0+2]*m[4*1+3]*m[4*3+0] - m[4*0+3]*m[4*1+0]*m[4*3+2] + m[4*0+0]*m[4*1+3]*m[4*3+2] + m[4*0+2]*m[4*1+0]*m[4*3+3] - m[4*0+0]*m[4*1+2]*m[4*3+3];
    a[4*1+3] = m[4*0+2]*m[4*1+3]*m[4*2+0] - m[4*0+3]*m[4*1+2]*m[4*2+0] + m[4*0+3]*m[4*1+0]*m[4*2+2] - m[4*0+0]*m[4*1+3]*m[4*2+2] - m[4*0+2]*m[4*1+0]*m[4*2+3] + m[4*0+0]*m[4*1+2]*m[4*2+3];
    a[4*2+0] = m[4*1+1]*m[4*2+3]*m[4*3+0] - m[4*1+3]*m[4*2+1]*m[4*3+0] + m[4*1+3]*m[4*2+0]*m[4*3+1] - m[4*1+0]*m[4*2+3]*m[4*3+1] - m[4*1+1]*m[4*2+0]*m[4*3+3] + m[4*1+0]*m[4*2+1]*m[4*3+3];
    a[4*2+1] = m[4*0+3]*m[4*2+1]*m[4*3+0] - m[4*0+1]*m[4*2+3]*m[4*3+0] - m[4*0+3]*m[4*2+0]*m[4*3+1] + m[4*0+0]*m[4*2+3]*m[4*3+1] + m[4*0+1]*m[4*2+0]*m[4*3+3] - m[4*0+0]*m[4*2+1]*m[4*3+3];
    a[4*2+2] = m[4*0+1]*m[4*1+3]*m[4*3+0] - m[4*0+3]*m[4*1+1]*m[4*3+0] + m[4*0+3]*m[4*1+0]*m[4*3+1] - m[4*0+0]*m[4*1+3]*m[4*3+1] - m[4*0+1]*m[4*1+0]*m[4*3+3] + m[4*0+0]*m[4*1+1]*m[4*3+3];
    a[4*2+3] = m[4*0+3]*m[4*1+1]*m[4*2+0] - m[4*0+1]*m[4*1+3]*m[4*2+0] - m[4*0+3]*m[4*1+0]*m[4*2+1] + m[4*0+0]*m[4*1+3]*m[4*2+1] + m[4*0+1]*m[4*1+0]*m[4*2+3] - m[4*0+0]*m[4*1+1]*m[4*2+3];
    a[4*3+0] = m[4*1+2]*m[4*2+1]*m[4*3+0] - m[4*1+1]*m[4*2+2]*m[4*3+0] - m[4*1+2]*m[4*2+0]*m[4*3+1] + m[4*1+0]*m[4*2+2]*m[4*3+1] + m[4*1+1]*m[4*2+0]*m[4*3+2] - m[4*1+0]*m[4*2+1]*m[4*3+2];
    a[4*3+1] = m[4*0+1]*m[4*2+2]*m[4*3+0] - m[4*0+2]*m[4*2+1]*m[4*3+0] + m[4*0+2]*m[4*2+0]*m[4*3+1] - m[4*0+0]*m[4*2+2]*m[4*3+1] - m[4*0+1]*m[4*2+0]*m[4*3+2] + m[4*0+0]*m[4*2+1]*m[4*3+2];
    a[4*3+2] = m[4*0+2]*m[4*1+1]*m[4*3+0] - m[4*0+1]*m[4*1+2]*m[4*3+0] - m[4*0+2]*m[4*1+0]*m[4*3+1] + m[4*0+0]*m[4*1+2]*m[4*3+1] + m[4*0+1]*m[4*1+0]*m[4*3+2] - m[4*0+0]*m[4*1+1]*m[4*3+2];
    a[4*3+3] = m[4*0+1]*m[4*1+2]*m[4*2+0] - m[4*0+2]*m[4*1+1]*m[4*2+0] + m[4*0+2]*m[4*1+0]*m[4*2+1] - m[4*0+0]*m[4*1+2]*m[4*2+1] - m[4*0+1]*m[4*1+0]*m[4*2+2] + m[4*0+0]*m[4*1+1]*m[4*2+2];


    det = matrix_determinant(m);

    for(i = 0; i < 16; ++i)
        a[i] /= det;

    matrix_identity(e);

    matrix_copy(b, m);
    matrix_multiply(b, a);

    matrix_copy(m, a);
}

void matrix_copy(float dest[16], float src[16])  {
    kdMemcpy(dest, src, 16*sizeof(float));
}

void matrix_print(float a[16])  {
    int i, j;
    KDchar tmp[10];
    KDchar buffer[50];

    for(i = 0; i < 4; ++i) {
        buffer[0] = '\0';
        for(j = 0; j < 4; ++j) {
            kdFtostr(tmp, 10, a[4*i+j]);
            kdStrncat_s(buffer, 50, tmp, 10);
            kdStrncat_s(buffer, 50, " ", 1);
        }
        kdLogMessage(buffer);
    }
}

void matrix_vector_multiply(float m[16],  float v[4])  {
    float res[4];
    int i, j;

    for(i = 0; i < 4; ++i)  {
        res[i] = 0;
        for(j = 0; j < 4; ++j)
            res[i] += m[i*4 + j] * v[j];
    }

    kdMemcpy(v, res, sizeof(res));
}

