
#include <stdlib.h>
#include <assert.h>
#include <stdio.h>
#include <string.h>
#include "GLES2/gl2.h"

#include "priv.h"
#include "nvtexfont2.h"

#include "helvetica.h"

/* ctype.h introduces non-portability with C libraries.                 */
/* Specifically, the symbol __ctype_b could be undefined.               */
/* We use ctype.h for the upper and lower case functions.               */
/* Avoid the problem altogether with some simple macros.                */
/* Note toupper() and tolower() assume the input is lower and upper     */
/* respectively, which it is as used below in getGlyphIndex().          */
#define myislower(c) ((c) >= 'a' && (c) <= 'z')
#define myisupper(c) ((c) >= 'A' && (c) <= 'Z')
#define mytoupper(c) ('A' + ((c)-'a'))
#define mytolower(c) ('a' + ((c)-'A'))

static int getGlyphIndex(NVTexfont2RasterFont * txf, int c)
{
    int indx;

    /* Automatically substitute uppercase letters with lowercase if not
       uppercase available (and vice versa). */
    if ((c >= txf->min_glyph) && (c < txf->min_glyph + txf->range)) {
        indx = txf->lut[c - txf->min_glyph];
        if (indx != -1) {
            return indx;
        }
        if (myislower(c)) {
            c = mytoupper(c);
            if ((c >= txf->min_glyph) && (c < txf->min_glyph + txf->range)) {
                return txf->lut[c - txf->min_glyph];
            }
        }
        if (myisupper(c)) {
            c = mytolower(c);
            if ((c >= txf->min_glyph) && (c < txf->min_glyph + txf->range)) {
                return txf->lut[c - txf->min_glyph];
            }
        }
    }
    return -1;
}


static unsigned char *getFontData(NVTexfont2RasterFontName font)
{
    switch (font) {
    case NV_TEXFONT_HELVETICA:
        return helvetica_font_data;
    default:
        return 0;
    }
}

/*
 * TexFont originally would fread() from a binary TexFont file.
 * NVTexfont2RasterFont has the TexFont data built into the library in the form of
 * a C header file.
 * So each "fread(..., file)" line in the original code has been replaced
 * by a "BUFREAD(..., buffer)" below, where BUFREAD is the following macro.
 */
#define BUFREAD(TO,SIZE,NUM,BUFFER)               \
        (memcpy(TO, BUFFER, SIZE*NUM), BUFFER += (SIZE*NUM), (NUM))

static NVTexfont2RasterFont *loadFontData(NVTexfont2RasterFontName font)
{
    NVTexfont2RasterFont *txf;
    GLfloat w, h, xstep, ystep;
    char fileid[4], tmp;
    unsigned char *texbitmap = NULL;
    int min_glyph, max_glyph;
    int endianness, swap, format, stride, width, height;
    int i, j, got;
    unsigned char *buffer = NULL;
    GLfloat *st;
    GLshort *vert;
    NVTexfont2GlyphInfo *tgi;

    txf = NULL;

    buffer = getFontData(font);
    if (!buffer) {
        lastError = "Unsupported font";
        goto error;
    }

    txf = (NVTexfont2RasterFont *) calloc(sizeof(NVTexfont2RasterFont), 1);
    if (txf == NULL) {
        lastError = "out of memory.";
        goto error;
    }
    /* For easy cleanup in error case. */
    txf->tgi = NULL;
    txf->lut = NULL;
    txf->teximage = NULL;

    got = BUFREAD(fileid, 1, 4, buffer);
    if (got != 4 || strncmp(fileid, "\377txf", 4)) {
        lastError = "not a texture font file.";
        goto error;
    }
    assert(sizeof(int) == 4);  /* Ensure external file format size. */
    got = BUFREAD(&endianness, sizeof(int), 1, buffer);
    if (got == 1 && endianness == 0x12345678) {
        swap = 0;
    } else if (got == 1 && endianness == 0x78563412) {
        swap = 1;
    } else {
        lastError = "not a texture font file.";
        goto error;
    }
#define EXPECT(n) if (got != n) { lastError = "premature end of file."; goto error; }
    got = BUFREAD(&format, sizeof(int), 1, buffer);
    EXPECT(1);
    got = BUFREAD(&txf->tex_width, sizeof(int), 1, buffer);
    EXPECT(1);
    got = BUFREAD(&txf->tex_height, sizeof(int), 1, buffer);
    EXPECT(1);
    got = BUFREAD(&txf->max_ascent, sizeof(int), 1, buffer);
    EXPECT(1);
    got = BUFREAD(&txf->max_descent, sizeof(int), 1, buffer);
    EXPECT(1);
    got = BUFREAD(&txf->num_glyphs, sizeof(int), 1, buffer);
    EXPECT(1);

    if (swap) {
        SWAPL(&format, tmp);
        SWAPL(&txf->tex_width, tmp);
        SWAPL(&txf->tex_height, tmp);
        SWAPL(&txf->max_ascent, tmp);
        SWAPL(&txf->max_descent, tmp);
        SWAPL(&txf->num_glyphs, tmp);
    }
    txf->tgi = (NVTexfont2GlyphInfo *)
        malloc(txf->num_glyphs * sizeof(NVTexfont2GlyphInfo));
    if (txf->tgi == NULL) {
        lastError = "out of memory.";
        goto error;
    }

    /* Ensure external file format size. */
    assert(sizeof(NVTexfont2GlyphInfo) == 12);

    got = BUFREAD(txf->tgi, sizeof(NVTexfont2GlyphInfo), txf->num_glyphs, buffer);
    EXPECT(txf->num_glyphs);

    if (swap) {
        for (i = 0; i < txf->num_glyphs; i++) {
            SWAPS(&txf->tgi[i].c, tmp);
            SWAPS(&txf->tgi[i].x, tmp);
            SWAPS(&txf->tgi[i].y, tmp);
        }
    }

    w = (GLfloat) txf->tex_width;
    h = (GLfloat) txf->tex_height;
    xstep = 0.5f / w;
    ystep = 0.5f / h;

    txf->st   = (GLfloat *) malloc(txf->num_glyphs * sizeof(GLfloat) * 8);
    txf->vert = (GLshort *) malloc(txf->num_glyphs * sizeof(GLshort) * 8);
    if(!txf->st || !txf->vert) {
        lastError = "out of memory.";
        goto error;
    }

    st   = txf->st;
    vert = txf->vert;
    tgi  = txf->tgi;
    for (i = 0; i < txf->num_glyphs; i++,tgi++,st+=8,vert+=8) {
        st  [0] = tgi->x / w - xstep;
        st  [1] = tgi->y / h - ystep;
        vert[0] = tgi->xoffset;
        vert[1] = tgi->yoffset;
        st  [2] = (tgi->x + tgi->width) / w + xstep;
        st  [3] = tgi->y / h - ystep;
        vert[2] = tgi->xoffset + tgi->width;
        vert[3] = tgi->yoffset;
        st  [4] = (tgi->x + tgi->width) / w + xstep;
        st  [5] = (tgi->y + tgi->height) / h + ystep;
        vert[4] = tgi->xoffset + tgi->width;
        vert[5] = tgi->yoffset + tgi->height;
        st  [6] = tgi->x / w - xstep;
        st  [7] = (tgi->y + tgi->height) / h + ystep;
        vert[6] = tgi->xoffset;
        vert[7] = tgi->yoffset + tgi->height;
    }

    min_glyph = txf->tgi[0].c;
    max_glyph = txf->tgi[0].c;
    for (i = 1; i < txf->num_glyphs; i++) {
        if (txf->tgi[i].c < min_glyph) {
            min_glyph = txf->tgi[i].c;
        }
        if (txf->tgi[i].c > max_glyph) {
            max_glyph = txf->tgi[i].c;
        }
    }
    txf->min_glyph = min_glyph;
    txf->range = max_glyph - min_glyph + 1;

    txf->lut = (int *) malloc(txf->range * sizeof(int));
    if (txf->lut == NULL) {
        lastError = "out of memory.";
        goto error;
    }
    for (i = 0; i < txf->range; i++) {
        txf->lut[i] = -1;
    }
    for (i = 0; i < txf->num_glyphs; i++) {
        txf->lut[txf->tgi[i].c - txf->min_glyph] = i;
    }

    switch (format) {
    case TXF_FORMAT_BYTE:
        txf->teximage = (unsigned char *)
            malloc(txf->tex_width * txf->tex_height);
        if (txf->teximage == NULL) {
            lastError = "out of memory.";
            goto error;
        }
        got = BUFREAD(txf->teximage, 1, txf->tex_width * txf->tex_height,
                      buffer);
        EXPECT(txf->tex_width * txf->tex_height);
        break;
    case TXF_FORMAT_BITMAP:
        width = txf->tex_width;
        height = txf->tex_height;
        stride = (width + 7) >> 3;
        texbitmap = (unsigned char *) malloc(stride * height);
        if (texbitmap == NULL) {
            lastError = "out of memory.";
            goto error;
        }
        got = BUFREAD(texbitmap, 1, stride * height, buffer);
        EXPECT(stride * height);
        txf->teximage = (unsigned char *) calloc(width * height, 1);
        if (txf->teximage == NULL) {
            lastError = "out of memory.";
            goto error;
        }
        for (i = 0; i < height; i++) {
            for (j = 0; j < width; j++) {
                if (texbitmap[i * stride + (j >> 3)] & (1 << (j & 7))) {
                    txf->teximage[i * width + j] = 255;
                }
            }
        }
        free(texbitmap);
        break;
    }

    return txf;

error:

    if (texbitmap) {
        free(texbitmap);
    }

    if (txf) {
        if (txf->tgi)
            free(txf->tgi);
        if (txf->lut)
            free(txf->lut);
        if (txf->teximage)
            free(txf->teximage);
        if (txf->st)
            free(txf->st);
        if (txf->vert)
            free(txf->vert);
        free(txf);
    }

    return NULL;
}


static int build2DMipmaps(GLenum target, GLint internalFormat,
                          GLsizei width, GLsizei height,
                          GLenum format, GLenum type, const void *data)
{
    int lod, x, y, c;
    int pitch=0, shift, components;
    int comp[4]; // up to 4 components in an OpenGL-ES external format
    GLubyte *buf, *dst;
    const GLubyte *src1, *src2;

    // Handle some error cases
    if ((width < 1 || height < 1) ||                 // 0 width/height
        ((width&(width-1) || height&(height-1))) ||  // power of two check
        (type != GL_UNSIGNED_BYTE)) {                // must be UNSIGNED_BYTE
        return -1;
    }

    switch (format) {
    case GL_RGBA:
    case GL_RGB:
    case GL_LUMINANCE_ALPHA: components = 2; break;
    case GL_LUMINANCE:
    case GL_ALPHA:           components = 1; break;
    default:                 components = 0; break;
    }

    // Reserve space for lod 1, (width/2) * (height/2) * 4, which is just...
    // For RGB or BGR, we reserve extra space.  So what?
    buf = malloc(width*height);
    if(!buf) return -1;

    lod = 0;
    do {
        // Download a level
        glTexImage2D(target, lod, internalFormat, width, height, 0,
                     format, GL_UNSIGNED_BYTE, data);

        // After the 1x1, we're done.
        if (width == 1 && height == 1) {
            break;
        }

        // Observation:
        // Normal MxN downsampling will average a 2x2 block of texels into 1.
        // Nx1 or 1xN downsampling will average 2 successive texels into 1.
        // The Nx1 and 1xN cases can be handled without distinction between
        // width and height.
        dst = buf;
        src1 =
        src2 =  data;
        if (width == 1 || height == 1) {
            // Nx1 or 1xN case
            shift = 1;
        } else {
            // General MxN case
            shift = 2;
            pitch = width*components;
            src2 += pitch;
        }

        if (width > 1)
            width >>= 1;
        if (height > 1)
            height >>= 1;

        for (y=0; y<height; y++) {
            for (x=0; x<width; x++) {
                for (c=0; c<components; c++) {
                    comp[c] = *src1++;
                }
                for (c=0; c<components; c++) {
                    comp[c] += *src1++;
                }
                if (shift == 2) {
                    for (c=0; c<components; c++) {
                        comp[c] += *src2++;
                    }
                    for (c=0; c<components; c++) {
                        comp[c] += *src2++;
                    }
                }

                // Write the downsampled texel.
                for (c=0; c<components; c++) {
                    *dst++ = comp[c]>>shift;
                }
            }

            // For MxN case, skip a row.
            if (shift == 2) {
                src1 += pitch;
                src2 += pitch;
            }
        }
        data = buf;
        lod++;
    } while (1);

    free(buf);
    return 0;
}




static int renderGlyph(NVTexfont2RasterFont * txf, int c)
{
    int indx = getGlyphIndex(txf, c);
    if (indx == -1)
        return 0;


    glDrawArrays(GL_TRIANGLE_FAN, 4*indx, 4);
    return txf->tgi[indx].advance;
}

static GLuint establishTexture(NVTexfont2RasterFont * txf,
                                 GLuint texobj,
                                 GLboolean setupMipmaps,
                                 GLenum minFilter,
                                 GLenum magFilter)
{
    GLint savedTextureBinding;

    if (txf->texobj == 0) {
        if (texobj == 0) {
            glGenTextures(1, &txf->texobj);
        } else {
            txf->texobj = texobj;
        }
    }

    glGetIntegerv(GL_TEXTURE_BINDING_2D, &savedTextureBinding);
    glBindTexture(GL_TEXTURE_2D, txf->texobj);

    if (setupMipmaps) {
        build2DMipmaps(GL_TEXTURE_2D, GL_ALPHA,
                          txf->tex_width, txf->tex_height,
                          GL_ALPHA, GL_UNSIGNED_BYTE, txf->teximage);
    } else {
        glTexImage2D(GL_TEXTURE_2D, 0, GL_ALPHA,
                     txf->tex_width, txf->tex_height, 0,
                     GL_ALPHA, GL_UNSIGNED_BYTE, txf->teximage);
    }

    glTexParameterf(GL_TEXTURE_2D,
                    GL_TEXTURE_MIN_FILTER, (GLfloat) minFilter);
    glTexParameterf(GL_TEXTURE_2D,
                    GL_TEXTURE_MAG_FILTER, (GLfloat) magFilter);

    glBindTexture(GL_TEXTURE_2D, savedTextureBinding);

    return txf->texobj;
}

int nvtexfont2InRasterFont(NVTexfont2RasterFont * txf, int c)
{
    /* NOTE: No uppercase/lowercase substituion. */
    if ((c >= txf->min_glyph) && (c < txf->min_glyph + txf->range)) {
        if (txf->lut[c - txf->min_glyph]) {
            return 1;
        }
    }
    return 0;
}


NVTexfont2RasterFont *nvtexfont2InitRasterFont(NVTexfont2RasterFontName font,
                                          GLuint        texobj,
                                          GLboolean     setupMipmaps,
                                          GLenum        minFilter,
                                          GLenum        magFilter)
{
    NVTexfont2RasterFont *txf = loadFontData(font);

    if (rasterProg == 0)
        initShaders();

    if (txf) {
        establishTexture(txf, texobj, setupMipmaps, minFilter, magFilter);
    }
    return txf;
}

void nvtexfont2UnloadRasterFont(NVTexfont2RasterFont * txf)
{

    deinitShaders();

    if (txf->texobj)
        glDeleteTextures(1, &txf->texobj);
    if (txf->teximage)
        free(txf->teximage);

    free(txf->tgi);
    free(txf->lut);
    free(txf->st);
    free(txf->vert);
    free(txf);
}

void nvtexfont2RenderString_All(NVTexfont2RasterFont * txf, char *string, float x, float y, float scale, float r, float g, float b)
{

    GLint savedActiveTexture;
    GLint savedTextureBinding;
    GLboolean savedBlendEnable, savedDepthTestEnabled;
    GLint savedBlendSrcRGB, savedBlendSrcAlpha, savedBlendDstRGB, savedBlendDstAlpha;
    GLint savedArrayBufferBinding;
    GLint savedVertArrayBinding, savedVertArrayEnabled, savedVertArraySize;
    GLint savedVertArrayStride, savedVertArrayType, savedVertArrayNormalized;
    void *savedVertArrayPointer;
    GLint savedTexArrayBinding, savedTexArrayEnabled, savedTexArraySize;
    GLint savedTexArrayStride, savedTexArrayType, savedTexArrayNormalized;
    void *savedTexArrayPointer;
    GLint savedProgram;

    GLuint uniScale, uniX, uniY, uniTex, uniColor, uniOffset;

    int offset = 0, i = 0;

    /*****************
     * SAVE GL STATE *
     *****************/

    glGetIntegerv(GL_ACTIVE_TEXTURE, &savedActiveTexture);
    glGetIntegerv(GL_TEXTURE_BINDING_2D, &savedTextureBinding);
    savedBlendEnable = glIsEnabled(GL_BLEND);
    savedDepthTestEnabled = glIsEnabled(GL_DEPTH_TEST);

    glGetIntegerv(GL_BLEND_SRC_RGB, &savedBlendSrcRGB);
    glGetIntegerv(GL_BLEND_SRC_ALPHA, &savedBlendSrcAlpha);
    glGetIntegerv(GL_BLEND_DST_RGB, &savedBlendDstRGB);
    glGetIntegerv(GL_BLEND_DST_ALPHA, &savedBlendDstAlpha);
    glGetIntegerv(GL_ARRAY_BUFFER_BINDING, &savedArrayBufferBinding);

    glGetVertexAttribiv(VERT_ARRAY, GL_VERTEX_ATTRIB_ARRAY_BUFFER_BINDING, &savedVertArrayBinding);
    glGetVertexAttribiv(VERT_ARRAY, GL_VERTEX_ATTRIB_ARRAY_ENABLED, &savedVertArrayEnabled);
    glGetVertexAttribiv(VERT_ARRAY, GL_VERTEX_ATTRIB_ARRAY_SIZE, &savedVertArraySize);
    glGetVertexAttribiv(VERT_ARRAY, GL_VERTEX_ATTRIB_ARRAY_STRIDE, &savedVertArrayStride);
    glGetVertexAttribiv(VERT_ARRAY, GL_VERTEX_ATTRIB_ARRAY_TYPE, &savedVertArrayType);
    glGetVertexAttribiv(VERT_ARRAY, GL_VERTEX_ATTRIB_ARRAY_NORMALIZED, &savedVertArrayNormalized);
    glGetVertexAttribPointerv(VERT_ARRAY, GL_VERTEX_ATTRIB_ARRAY_POINTER, &savedVertArrayPointer);

    glGetVertexAttribiv(TEX_ARRAY, GL_VERTEX_ATTRIB_ARRAY_BUFFER_BINDING, &savedTexArrayBinding);
    glGetVertexAttribiv(TEX_ARRAY, GL_VERTEX_ATTRIB_ARRAY_ENABLED, &savedTexArrayEnabled);
    glGetVertexAttribiv(TEX_ARRAY, GL_VERTEX_ATTRIB_ARRAY_SIZE, &savedTexArraySize);
    glGetVertexAttribiv(TEX_ARRAY, GL_VERTEX_ATTRIB_ARRAY_STRIDE, &savedTexArrayStride);
    glGetVertexAttribiv(TEX_ARRAY, GL_VERTEX_ATTRIB_ARRAY_TYPE, &savedTexArrayType);
    glGetVertexAttribiv(TEX_ARRAY, GL_VERTEX_ATTRIB_ARRAY_NORMALIZED, &savedTexArrayNormalized);
    glGetVertexAttribPointerv(TEX_ARRAY, GL_VERTEX_ATTRIB_ARRAY_POINTER, &savedTexArrayPointer);
    glGetIntegerv(GL_CURRENT_PROGRAM, &savedProgram);

    /************
     * SETUP GL *
     ************/

    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, txf->texobj);
    glEnable(GL_BLEND);
    glDisable(GL_DEPTH_TEST);
    glBlendFuncSeparate(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA, GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glBindBuffer(GL_ARRAY_BUFFER, 0);

    glEnableVertexAttribArray(VERT_ARRAY);
    glVertexAttribPointer(VERT_ARRAY, 2, GL_SHORT, GL_FALSE, 0, txf->vert);
    glEnableVertexAttribArray(TEX_ARRAY);
    glVertexAttribPointer(TEX_ARRAY, 2, GL_FLOAT, GL_FALSE, 0, txf->st);

    /*****************
     * SETUP SHADERS *
     *****************/

    glUseProgram(rasterProg);
    uniScale = glGetUniformLocation(rasterProg, "scale");
    uniX = glGetUniformLocation(rasterProg, "screenX");
    uniY = glGetUniformLocation(rasterProg, "screenY");
    uniTex = glGetUniformLocation(rasterProg, "fontTex");
    uniColor = glGetUniformLocation(rasterProg, "color");

    // Adjust scale so that character size is approximately 1/10th screen height
    glUniform1f(uniScale, scale * 0.2f / txf->max_ascent);
    glUniform1f(uniX, x);
    glUniform1f(uniY, y);
    glUniform1i(uniTex, 0);
    glUniform3f(uniColor, r, g, b);

    uniOffset = glGetUniformLocation(rasterProg, "offset");

    /*************
     * DRAW TEXT *
     *************/

    while (string[i]) {
        glUniform1i(uniOffset, offset);
        offset += renderGlyph(txf, string[i++]);
    }

    /********************
     * RESTORE GL STATE *
     ********************/

    glBindBuffer(GL_ARRAY_BUFFER, savedArrayBufferBinding);

    if (!savedVertArrayEnabled)
        glDisableVertexAttribArray(VERT_ARRAY);
    if (!savedTexArrayEnabled)
        glDisableVertexAttribArray(TEX_ARRAY);

    glVertexAttribPointer(VERT_ARRAY, savedVertArraySize, savedVertArrayType, savedVertArrayNormalized,
                          savedVertArrayStride, savedVertArrayPointer);
    glVertexAttribPointer(TEX_ARRAY, savedTexArraySize, savedTexArrayType, savedTexArrayNormalized,
                          savedTexArrayStride, savedTexArrayPointer);

    if (!savedBlendEnable)
        glDisable(GL_BLEND);
    glBlendFuncSeparate(savedBlendSrcRGB, savedBlendDstRGB, savedBlendSrcAlpha, savedBlendDstAlpha);

    glUseProgram(savedProgram);

    glActiveTexture(savedActiveTexture);
    glBindTexture(GL_TEXTURE_2D, savedTextureBinding);


    if (savedDepthTestEnabled)
        glEnable(GL_DEPTH_TEST);
}


void nvtexfont2RenderString_Pos(NVTexfont2RasterFont * txf, NVTexfont2Context tfc,
                                char *string,
                                float x, float y)
{
    nvtexfont2RenderString_All(txf, string, x, y, ((nvtexfont2State *)tfc)->scale, ((nvtexfont2State *)tfc)->r, ((nvtexfont2State *)tfc)->g, ((nvtexfont2State *)tfc)->b);
}

void nvtexfont2RenderString(NVTexfont2RasterFont * txf, NVTexfont2Context tfc,
                            char *string)
{
    nvtexfont2RenderString_All(txf, string, ((nvtexfont2State *)tfc)->x, ((nvtexfont2State *)tfc)->y, ((nvtexfont2State *)tfc)->scale, ((nvtexfont2State *)tfc)->r, ((nvtexfont2State *)tfc)->g, ((nvtexfont2State *)tfc)->b);
}


void nvtexfont2GetStringMetrics(NVTexfont2RasterFont * txf, char *string, int len,
                               int *width, int *max_ascent, int *max_descent)
{
    int w, i, indx;

    w = 0;
    for (i = 0; i < len; i++) {
        if (string[i] == 27) {
            switch (string[i + 1]) {
            case 'M': i += 4;  break;
            case 'T': i += 7;  break;
            case 'L': i += 7;  break;
            case 'F': i += 13; break;
            }
        } else {
            indx = getGlyphIndex(txf, string[i]);
            if (indx != -1) {
                w += txf->tgi[indx].advance;
            }
        }
    }
    *width = w;
    *max_ascent = txf->max_ascent;
    *max_descent = txf->max_descent;
}
