#include <KD/kd.h>
#include "shaders.h"
#include "nvgl2demo_common.h"

// When using binary shaders, we compile them in so the application
//   doesn't need to load extra data files
#ifdef USE_BINARY_SHADERS
static const char shad_bubbleVert[] = {
#   include "bubble_vert.cghex"
};
static const char shad_bubbleFrag[] = {
#   include "bubble_frag.cghex"
};
static const char shad_meshVert[] = {
#   include "mesh_vert.cghex"
};
static const char shad_meshFrag[] = {
#   include "mesh_frag.cghex"
};
static const char shad_cubeVert[] = {
#   include "envCube_vert.cghex"
};
static const char shad_cubeFrag[] = {
#   include "envCube_frag.cghex"
};
static const char shad_mouseVert[] = {
#   include "mouse_vert.cghex"
};
static const char shad_mouseFrag[] = {
#   include "mouse_frag.cghex"
};
#endif

// Reflective bubble shader
GLint prog_bubble = 0;
GLint uloc_bubbleViewMat;
GLint uloc_bubbleNormMat;
GLint uloc_bubbleProjMat;
GLint uloc_bubbleTexUnit;
GLint aloc_bubbleVertex;
GLint aloc_bubbleNormal;

// Wireframe/point mesh shader
GLint prog_mesh = 0;
GLint uloc_meshViewMat;
GLint uloc_meshProjMat;
GLint aloc_meshVertex;

// Environment cube shader
GLint prog_cube = 0;
GLint uloc_cubeProjMat;
GLint uloc_cubeTexUnit;
GLint aloc_cubeVertex;

// Crosshair shader
GLint prog_mouse = 0;
GLint uloc_mouseWindow;
GLint uloc_mouseCenter;
GLint aloc_mouseVertex;

int LoadShaders(void)
{
    GLboolean success;

#   ifdef USE_BINARY_SHADERS
    // If using binary shaders, create the programs from the hardcoded headers
    prog_bubble =
        NvGl2DemoLoadShaderBinaries(
            shad_bubbleVert, sizeof(shad_bubbleVert),
            shad_bubbleFrag, sizeof(shad_bubbleFrag),
            GL_FALSE);
    prog_mesh =
        NvGl2DemoLoadShaderBinaries(
            shad_meshVert, sizeof(shad_meshVert),
            shad_meshFrag, sizeof(shad_meshFrag),
            GL_FALSE);
    prog_cube =
        NvGl2DemoLoadShaderBinaries(
            shad_cubeVert, sizeof(shad_cubeVert),
            shad_cubeFrag, sizeof(shad_cubeFrag),
            GL_FALSE);
    prog_mouse =
        NvGl2DemoLoadShaderBinaries(
            shad_mouseVert, sizeof(shad_mouseVert),
            shad_mouseFrag, sizeof(shad_mouseFrag),
            GL_FALSE);
#   else
    // Otherwise, load and compile the shader sources
    prog_bubble =
        NvGl2DemoLoadShaderSources(
            "bubble_vert.glslv", "bubble_frag.glslf", GL_FALSE);
    prog_mesh =
        NvGl2DemoLoadShaderSources(
            "mesh_vert.glslv", "mesh_frag.glslf", GL_FALSE);
    prog_cube =
        NvGl2DemoLoadShaderSources(
            "envCube_vert.glslv", "envCube_frag.glslf", GL_FALSE);
    prog_mouse =
        NvGl2DemoLoadShaderSources(
            "mouse_vert.glslv", "mouse_frag.glslf", GL_FALSE);
#   endif

    success = prog_bubble && prog_mesh && prog_cube && prog_mouse;
    if (!success) {
        //NvGl2DemoLogMessage("Error occured loading shaders");
        return 0;
    }

    // Load locations for bubble shader
    uloc_bubbleViewMat = glGetUniformLocation(prog_bubble, "modelview");
    uloc_bubbleNormMat = glGetUniformLocation(prog_bubble, "imodelview");
    uloc_bubbleProjMat = glGetUniformLocation(prog_bubble, "projection");
    uloc_bubbleTexUnit = glGetUniformLocation(prog_bubble, "texunit");
    aloc_bubbleVertex  = glGetAttribLocation(prog_bubble, "vertex");
    aloc_bubbleNormal  = glGetAttribLocation(prog_bubble, "normal");
    success =  (uloc_bubbleViewMat >= 0)
            && (uloc_bubbleNormMat >= 0)
            && (uloc_bubbleProjMat >= 0)
            && (uloc_bubbleTexUnit >= 0)
            && (aloc_bubbleVertex >= 0)
            && (aloc_bubbleNormal >= 0);
    if (!success) {
        //NvGl2DemoLogMessage("Error occured retrieving bubble shader locations");
        return 0;
    }

    // Load locations for environment cube shader
    uloc_meshViewMat = glGetUniformLocation(prog_mesh, "modelview");
    uloc_meshProjMat = glGetUniformLocation(prog_mesh, "projection");
    aloc_meshVertex =  glGetAttribLocation(prog_mesh, "vertex");
    success =  (uloc_meshViewMat >= 0)
            && (uloc_meshProjMat >= 0)
            && (aloc_meshVertex >= 0);
    if (!success) {
        //NvGl2DemoLogMessage("Error occured retrieving mesh shader locations");
        return 0;
    }

    // Load locations for environement cube shader
    uloc_cubeProjMat = glGetUniformLocation(prog_cube, "projection");
    uloc_cubeTexUnit = glGetUniformLocation(prog_cube, "texunit");
    aloc_cubeVertex =  glGetAttribLocation(prog_cube, "vertex");
    success =  (uloc_cubeProjMat >= 0)
            && (uloc_cubeTexUnit >= 0)
            && (aloc_cubeVertex >= 0);
    if (!success) {
        //NvGl2DemoLogMessage("Error occured retrieving cube shader locations");
        return 0;
    }

    // Load locations for environement mouse shader
    uloc_mouseWindow = glGetUniformLocation(prog_mouse, "window");
    uloc_mouseCenter = glGetUniformLocation(prog_mouse, "center");
    aloc_mouseVertex = glGetAttribLocation(prog_mouse, "vertex");
    success =  (uloc_mouseWindow >= 0)
            && (uloc_mouseCenter >= 0)
            && (aloc_mouseVertex >= 0);
    if (!success) {
        //NvGl2DemoLogMessage("Error occured retrieving mouse shader locations");
        return 0;
    }

    return 1;
}

void FreeShaders(void)
{
    if (prog_mouse)  glDeleteProgram(prog_mouse);
    if (prog_cube)   glDeleteProgram(prog_cube);
    if (prog_mesh)   glDeleteProgram(prog_mesh);
    if (prog_bubble) glDeleteProgram(prog_bubble);
}
